@@ -39,10 +39,12 @@ typedef struct {
 	GNOME_Speech_speech_callback_type type;
 	gint text_id;
 	gint offset;
+	gint id;
 } index_queue_entry;
 
 
 static gint text_id = 0;
+static gint marker_id = 0;
 
 
 static GObjectClass *parent_class;
@@ -97,9 +99,46 @@ dectalk_synthesis_driver_from_servant (PortableServer_Servant *servant)
 static gboolean
 dectalk_idle (gpointer data)
 {
+	gchar  *converted_text;
+	GError *gerror = NULL;
+	
 	DectalkSynthesisDriver *d = DECTALK_SYNTHESIS_DRIVER (data);
 
-	TextToSpeechSpeak (d->handle, d->text, TTS_FORCE);
+	/* After some conversation with the DECtalk engineers, it is
+	 * unclear about whether DECtalk wants strings from code page 
+	 * 850 or ISO8859-1.  Apparently there are notes in the DECtalk
+	 * sources that talk about code page 850, but the general 
+	 * consensus seems to be that ISO8859-1 is really what is 
+	 * expected.  
+	 *
+	 * We're getting strings in UTF-8, so we convert.  This will 
+	 * help prevent DECtalk from crashing (see bug 346357 at
+	 * http://bugzilla.gnome.org/show_bug.cgi?id=347357).  In
+	 * addition, since DECtalk has been shown to crash if we give
+	 * it bad strings, we will not send anything to speak if the
+	 * conversion failed.
+	 */
+	converted_text = g_convert_with_fallback (
+	        d->text, 
+		-1,          /* d->text is a null-terminated string */
+		"ISO8859-1", /* what DECtalk wants                  */
+		"UTF-8",     /* what we are getting                 */
+		" ",         /* the fallback hint                   */
+		NULL, NULL,  /* bytes read and written              */
+		&gerror);    /* what went wrong, if anything        */
+
+	if (!converted_text) {
+	        printf ("STRING CONVERSION FAILED: (string=%s) (error=%s)!\n", 
+		        d->text, gerror->message);
+        } else {
+#if 0
+	        printf ("STRING CONVERSION: (from=%s)\n", d->text); 
+	        printf ("STRING CONVERSION:   (to=%s)\n", converted_text); 
+#endif
+		TextToSpeechSpeak (d->handle, converted_text, TTS_FORCE);
+		g_free (converted_text);
+        }
+
 	g_free (d->text);
 	d->text = NULL;
 	d->text_idle = 0;
@@ -130,11 +169,13 @@ dectalk_synthesis_driver_add_index(DectalkSynthesisDriver *d,
 				   GNOME_Speech_SpeechCallback cb,
 				   GNOME_Speech_speech_callback_type type,
 				   gint text_id,
-				   gint offset)
+				   gint offset,
+				   gint marker_id)
 {
 	index_queue_entry *e;
 	CORBA_Environment ev;
-	
+	gchar *marker;
+
 	e = g_new(index_queue_entry, 1);
 	CORBA_exception_init (&ev);
 	e->cb = CORBA_Object_duplicate (cb, &ev);
@@ -142,12 +183,15 @@ dectalk_synthesis_driver_add_index(DectalkSynthesisDriver *d,
 	e->type = type;
 	e->text_id = text_id;
 	e->offset = offset;
+	e->id = marker_id;
 
 	g_mutex_lock (d->mutex);
 	d->index_queue = g_slist_append(d->index_queue, e);
 	g_mutex_unlock (d->mutex);
-	
-	dectalk_add_string (d, " [:index mark 00]");
+
+	marker = g_strdup_printf (" [:index mark %d]", marker_id);
+	dectalk_add_string (d, marker);
+	g_free (marker);
 }
 
 
@@ -157,8 +201,8 @@ dectalk_timeout (void *data)
 	DectalkSynthesisDriver *d = DECTALK_SYNTHESIS_DRIVER (data);
 	CORBA_Environment ev;
 	index_queue_entry *e;
-	
-	if (d->post_queue) {
+
+	while (d->post_queue) {
 		g_mutex_lock (d->mutex);
 		e = (index_queue_entry *) d->post_queue->data;
 		d->post_queue = g_slist_remove_link (d->post_queue, d->post_queue);
@@ -185,16 +229,35 @@ dectalk_callback (LONG lParam1,
 {
 	DectalkSynthesisDriver *d = DECTALK_SYNTHESIS_DRIVER(dwCallbackParameter);
 	index_queue_entry *e = NULL;
-	
+
 	switch (uiMsg) {
 	case TTS_MSG_INDEX_MARK:
 		if (d && d->index_queue) {
 			g_mutex_lock (d->mutex);
-			e = (index_queue_entry *) d->index_queue->data;
-			d->index_queue = g_slist_remove_link (d->index_queue, d->index_queue);
-			d->post_queue = g_slist_append (d->post_queue, e);
+			do {
+			    e = (index_queue_entry *) d->index_queue->data;
+			    if (lParam2 >= e->id)
+			    {
+				d->index_queue = g_slist_remove_link (d->index_queue, d->index_queue);
+				d->post_queue = g_slist_append (d->post_queue, e);
+			    }
+			}while (lParam2 >= e->id && d->index_queue);
 			g_mutex_unlock (d->mutex);
-		}
+		};
+		break;
+	case TTS_MSG_STATUS:
+		if (lParam1 == TTS_AUDIO_PLAY_STOP) {
+			g_mutex_lock (d->mutex);
+			while (d->index_queue) {
+			    e = (index_queue_entry *) d->index_queue->data;
+			    d->index_queue = g_slist_remove_link (d->index_queue, d->index_queue);
+			    d->post_queue = g_slist_append (d->post_queue, e);
+			    if (e->type == GNOME_Speech_speech_callback_speech_ended)
+				break;
+			}
+			g_mutex_unlock (d->mutex);
+		};
+		break;
 	}
 	return;
 }
@@ -392,10 +455,9 @@ dectalk_driverInit (PortableServer_Servant servant,
 		    CORBA_Environment *ev)
 {
 	DectalkSynthesisDriver *d = dectalk_synthesis_driver_from_servant (servant);
-	UINT devOptions = 0;
+	UINT devOptions = WAVE_OPEN_SHAREABLE;
 	int devNo = (int) WAVE_MAPPER;
 
-
 	if (TextToSpeechStartup(&d->handle, devNo, devOptions,
 				dectalk_callback, (gint) d) != MMSYSERR_NOERROR) {
 		d->handle = NULL;
@@ -412,6 +474,7 @@ dectalk_driverInit (PortableServer_Servant servant,
 						    dectalk_timeout, d, NULL);
 		d->initialized = TRUE;
 	}
+
 	return d->initialized;
 }
 
@@ -497,7 +560,21 @@ dectalk_synthesis_driver_finalize(GObject *obj)
 	DectalkSynthesisDriver *d = DECTALK_SYNTHESIS_DRIVER(obj);
 	
 	if (d->handle != NULL) {
+#if 0
+	  /* [[[WDW: Commented this out for gnome-speech 0.3.10 as
+	     a workaround for the problem where DECtalk can suddenly
+	     go silent and we cannot detect it.  When we try to exit
+	     the driver gracefully afterwards, the driver will hang
+	     on this call.  The impact of commenting this out is that
+	     any speech in progress will stop when the driver exits.
+	     Oh well.  At least it will exit and not hang.  Note that
+	     we will also be working with Fonix to determine the
+	     underlying cause of why DECtalk is suddenly going silent.
+	     If we can solve that problem, this line probably can be
+	     re-added.
+	  */
 		TextToSpeechSync (d->handle);
+#endif
 		TextToSpeechShutdown(d->handle);
 	}
 	if (d->timeout_id > 0)
@@ -554,9 +631,11 @@ dectalk_synthesis_driver_say (DectalkSynthesisDriver *d,
 	Speaker *speaker = SPEAKER (s);
 	gchar *beginning = text;
 	gchar *end, *word;
-	
-        /* If this speaker wasn't the last one to speak, reset the speech parameters */
 
+
+	g_assert (g_utf8_validate (text, -1, NULL));
+
+        /* If this speaker wasn't the last one to speak, reset the speech parameters */
 	if (d->last_speaker != s || speaker_needs_parameter_refresh(speaker)) {
 		dectalk_add_string (d, (char *) s->voice);
 		speaker_refresh_parameters (speaker);
@@ -568,29 +647,31 @@ dectalk_synthesis_driver_say (DectalkSynthesisDriver *d,
 		dectalk_synthesis_driver_add_index(d,
 						   speaker->cb,
 						   GNOME_Speech_speech_callback_speech_started,
-						   text_id, 0);
+						   text_id, 0, marker_id++);
 		
 		/* Add index between words */
 		
 		while (*beginning) {
 			end = beginning;
-			while (*end && !g_ascii_isspace(*end))
-				end++;
+			while (*end && g_unichar_isspace(g_utf8_get_char (end)))
+				end = beginning = g_utf8_next_char (end);
+			while (*end && !g_unichar_isspace(g_utf8_get_char (end)))
+				end = g_utf8_next_char (end);
 			word = g_strndup (beginning, end-beginning);
 			dectalk_add_string (d, word);
 			g_free (word);
 			dectalk_synthesis_driver_add_index(d,
 						   speaker->cb,
 						   GNOME_Speech_speech_callback_speech_progress,
-						   text_id, end-text);
+						   text_id, end-text, marker_id++);
 			if (!*end)
 				break;
-			beginning = end+1;
+			beginning = g_utf8_next_char (end);
 		}
 		dectalk_synthesis_driver_add_index(d,
 						   speaker->cb,
 						   GNOME_Speech_speech_callback_speech_ended,
-						   text_id, 0);
+						   text_id, 0, marker_id++);
 	}
 	else
 		dectalk_add_string (d, (char *) text);
@@ -636,7 +717,7 @@ dectalk_synthesis_driver_wait (DectalkSynthesisDriver *d)
 gboolean
 dectalk_synthesis_driver_is_speaking (DectalkSynthesisDriver *d)
 {
-	gint lparam1, lparam2;
+	DWORD lparam1, lparam2;
 
 	TextToSpeechGetStatus (d->handle, &lparam1, &lparam2, STATUS_SPEAKING);
 	return TRUE;
