@@ -45,6 +45,53 @@ typedef struct {
 static gint text_id = 0;
 static GObjectClass *parent_class;
 
+typedef struct _eciLocale {
+	char *pLocale;
+	enum ECILanguageDialect langID;
+} eciLocale, *eciLocaleList;
+
+static eciLocale eciLocales[] = {
+	{"en_US", eciGeneralAmericanEnglish },
+	{"en_GB", eciBritishEnglish },
+	{"es_ES", eciCastilianSpanish},
+	{"es_MX", eciMexicanSpanish },
+	{"fr_FR", eciStandardFrench },
+	{"ca_FR", eciCanadianFrench },
+	{"de_DE", eciStandardGerman },
+	{"it_IT", eciStandardItalian},
+	{"zh_CN", eciMandarinChinese},
+	{"zh_CN", eciMandarinChineseGB },
+	{"zh_CN", eciMandarinChinesePinYin},
+	{"zh_CN", eciMandarinChineseUCS},
+	{"zh_TW", eciTaiwaneseMandarin },
+	{"zh_TW", eciTaiwaneseMandarinBig5},
+	{"zh_TW", eciTaiwaneseMandarinZhuYin},
+	{"zh_TW", eciTaiwaneseMandarinPinYin},
+	{"zh_TW", eciTaiwaneseMandarinUCS },
+	{"pt_BR", eciBrazilianPortuguese},
+	{"ja_JP", eciStandardJapanese},
+	{"ja_JP", eciStandardJapaneseSJIS },
+	{"ja_JP", eciStandardJapaneseUCS},
+	{"fi_FI", eciStandardFinnish},
+	{"ko_KR", eciStandardKorean },
+	{"ko_KR", eciStandardKoreanUHC },
+	{"ko_KR", eciStandardKoreanUCS },
+	{"zh_HK", eciStandardCantonese },
+	{"zh_HK", eciStandardCantoneseGB},
+	{"zh_HK", eciStandardCantoneseUCS },
+	{"zh_HK", eciHongKongCantonese },
+	{"zh_HK", eciHongKongCantoneseBig5},
+	{"zh_HK", eciHongKongCantoneseUCS },
+	{"nl_BE", eciStandardDutch},
+	{"no_NO", eciStandardNorwegian },
+	{"sv_SE", eciStandardSwedish},
+	{"da_DK", eciStandardDanish },
+	{"en_US", eciStandardReserved},
+	{"th_TH", eciStandardThai },
+	{"th_TH", eciStandardThaiTIS},
+	{NULL, 0}
+};
+
 /*
  *
  * index_queue_entry functions
@@ -237,33 +284,49 @@ viavoice_isInitialized (PortableServer_Servant servant,
 
 
 static GSList *
-get_voice_list (void)
+get_voice_list (GNOME_Speech_VoiceInfo *vs)
 {
 	GSList *l = NULL;
 	ECIHand tmp_handle;
-	int i;
+	int h, i;
+	enum ECILanguageDialect origLang;
   
 	tmp_handle = eciNew ();
 	if (tmp_handle == NULL_ECI_HAND)
 		return NULL;
 
-	for (i = 1; i <= ECI_PRESET_VOICES; i++) {
-		GNOME_Speech_VoiceInfo *info;
-		char tmp[ECI_VOICE_NAME_LENGTH + 1];
-		int gender;
-    
-		info = GNOME_Speech_VoiceInfo__alloc ();
-		eciGetVoiceName (tmp_handle, i, tmp);
-		info->language = CORBA_string_dup("en_US");
-		info->name = CORBA_string_dup (tmp);
-		gender = eciGetVoiceParam (tmp_handle, i, eciGender);
-		if (gender == 0)
-			info->gender = GNOME_Speech_gender_male;
-		else if (gender == 1)
-			info->gender = GNOME_Speech_gender_female;
-		l = g_slist_prepend (l, info);
+	origLang = eciGetParam(tmp_handle, eciLanguageDialect);
+
+	for (h = 0; NULL != eciLocales[h].pLocale; h++) {
+
+		if (vs && (0 != strcmp(vs->language, eciLocales[h].pLocale))) {
+			continue;
+		}
+
+		if (eciSetParam(tmp_handle, eciLanguageDialect,
+		eciLocales[h].langID) < 0) {
+			continue;
+		}
+
+		for (i = 1; i <= ECI_PRESET_VOICES; i++) {
+			GNOME_Speech_VoiceInfo *info;
+			char tmp[ECI_VOICE_NAME_LENGTH + 1];
+			int gender;
+	    
+			info = GNOME_Speech_VoiceInfo__alloc ();
+			eciGetVoiceName (tmp_handle, i, tmp);
+			info->language= CORBA_string_dup(eciLocales[h].pLocale);
+			info->name = CORBA_string_dup (tmp);
+			gender = eciGetVoiceParam (tmp_handle, i, eciGender);
+			if (gender == 0)
+				info->gender = GNOME_Speech_gender_male;
+			else if (gender == 1)
+				info->gender = GNOME_Speech_gender_female;
+			l = g_slist_prepend (l, info);
+		}
 	}
 	l = g_slist_reverse (l);
+	eciSetParam(tmp_handle, eciLanguageDialect, origLang);
 	eciDelete (tmp_handle);
 	return l;
 }
@@ -321,7 +384,7 @@ viavoice_getVoices (PortableServer_Servant servant,
 	GNOME_Speech_VoiceInfoList *rv;
 	GSList *l;
 
-	l = get_voice_list ();
+	l = get_voice_list (voice_spec);
 	rv = voice_info_list_from_voice_list (l);
 	voice_list_free (l);
 	return rv;
@@ -336,7 +399,7 @@ viavoice_getAllVoices (PortableServer_Servant servant,
 	GNOME_Speech_VoiceInfoList *rv;
 	GSList *l;
 
-	l = get_voice_list ();
+	l = get_voice_list (NULL);
 	rv = voice_info_list_from_voice_list (l);
 	voice_list_free (l);
 	return rv;
@@ -505,9 +568,14 @@ viavoice_synthesis_driver_say (ViavoiceSynthesisDriver *d,
 			       gchar *text)
 {
 	Speaker *speaker = SPEAKER (s);
-	gchar *beginning = text;
-	gchar *end, *word;
-	
+	gchar *end, *word, *converted_text;
+        GError *error=NULL;
+
+	converted_text=g_convert_with_fallback(text, -1, "ISO-8859-1", "UTF-8", " ", NULL, NULL, &error);
+	gchar *beginning = converted_text;
+		
+
+
         /* If this speaker wasn't the last one to speak, reset the speech parameters */
 
 	if (d->last_speaker != s || speaker_needs_parameter_refresh(speaker)) {
@@ -516,6 +584,7 @@ viavoice_synthesis_driver_say (ViavoiceSynthesisDriver *d,
 	}
 	
 	text_id++;
+      if (converted_text != NULL) {
 	if (speaker->cb != CORBA_OBJECT_NIL) {
 		viavoice_synthesis_driver_add_index(d,
 						   speaker->cb,
@@ -545,11 +614,13 @@ viavoice_synthesis_driver_say (ViavoiceSynthesisDriver *d,
 						   text_id, 0);
 	}
 	else
-		eciAddText (d->handle, (char *) text);
+		eciAddText (d->handle, (char *) converted_text);
 	if (eciSynthesize (d->handle))
 		return text_id;
 	else
 		return -1;
+ }
+else return -1;
 }
 
 
@@ -584,6 +655,14 @@ viavoice_synthesis_driver_set_voice_param (const ViavoiceSynthesisDriver *d,
 	eciSetVoiceParam (d->handle, 0, param, new_value);
 }
 
+void
+viavoice_synthesis_driver_set_device_param (const ViavoiceSynthesisDriver *d,
+					   enum ECIParam param,
+					   gint new_value)
+{
+	eciSetParam (d->handle, param, new_value);
+}
+
 
 gboolean
 viavoice_synthesis_driver_set_voice (const ViavoiceSynthesisDriver *d,
@@ -593,3 +672,29 @@ viavoice_synthesis_driver_set_voice (const ViavoiceSynthesisDriver *d,
 		      (int) v,
 		      0);
 }
+
+char *
+dialect_to_language(enum ECILanguageDialect langID) {
+	int h;
+
+	for (h = 0; NULL != eciLocales[h].pLocale; h++) {
+
+		if (langID == eciLocales[h].langID) {
+			return  eciLocales[h].pLocale;
+		}
+	}
+}
+
+enum ECILanguageDialect
+language_to_dialect(char *lang) {
+	int h;
+
+	for (h = 0; NULL != eciLocales[h].pLocale; h++) {
+
+		if (0 != strcmp(lang, eciLocales[h].pLocale)) {
+			return eciLocales[h].langID;
+		}
+	}
+
+	return eciGeneralAmericanEnglish;
+}
